# -*- coding: utf-8 -*-

"""Validation tests for :mod:`biomappings`."""

import itertools as itt
import unittest
from collections import defaultdict

import bioregistry

from biomappings import (
    load_false_mappings,
    load_mappings,
    load_predictions,
    load_unsure,
)
from biomappings.resources import (
    MappingTuple,
    PredictionTuple,
    load_curators,
    mapping_sort_key,
)
from biomappings.utils import (
    InvalidIdentifierPattern,
    InvalidNormIdentifier,
    check_valid_prefix_id,
    get_canonical_tuple,
)

mappings = load_mappings()
predictions = load_predictions()
incorrect = load_false_mappings()
unsure = load_unsure()


def _iter_groups():
    for group, label in [
        (mappings, "positive"),
        (incorrect, "negative"),
        (predictions, "predictions"),
        (unsure, "unsure"),
    ]:
        for i, mapping in enumerate(group, start=2):
            yield label, i, mapping


class TestIntegrity(unittest.TestCase):
    """Data integrity tests."""

    def test_canonical_prefixes(self):
        """Test that all mappings use canonical bioregistry prefixes."""
        valid_prefixes = set(bioregistry.read_registry())
        for label, line, mapping in _iter_groups():
            source_prefix, target_prefix = mapping["source prefix"], mapping["target prefix"]
            self.assertIn(
                source_prefix,
                valid_prefixes,
                msg=f"Invalid prefix: {source_prefix} on {label}:{line}",
            )
            self.assertIn(
                target_prefix,
                valid_prefixes,
                msg=f"Invalid prefix: {target_prefix} on {label}:{line}",
            )

    def test_normalized_identifiers(self):
        """Test that all identifiers have been normalized (based on bioregistry definition)."""
        for label, line, mapping in _iter_groups():
            self.assert_canonical_identifier(
                mapping["source prefix"], mapping["source identifier"], label, line
            )
            self.assert_canonical_identifier(
                mapping["target prefix"], mapping["target identifier"], label, line
            )

    def assert_canonical_identifier(
        self, prefix: str, identifier: str, label: str, line: int
    ) -> None:
        """Assert a given identifier is canonical.

        :param prefix: The prefix to check
        :param identifier: The identifier in the semantic space for the prefix
        :param label: The label of the mapping file
        :param line: The line number of the mapping
        """
        try:
            check_valid_prefix_id(prefix, identifier)
        except InvalidNormIdentifier as e:
            self.fail(f"[{label}:{line}] {e}")
        except InvalidIdentifierPattern as e:
            self.fail(f"[{label}:{line}] {e}")

    def test_contributors(self):
        """Test all contributors have an entry in the curators.tsv file."""
        contributor_orcids = {row["orcid"] for row in load_curators()}
        for mapping in itt.chain(mappings, incorrect, unsure):
            source = mapping["source"]
            if not source.startswith("orcid:"):
                continue
            self.assertIn(source[len("orcid:") :], contributor_orcids)


def _extract_redundant(counter):
    return [(key, values) for key, values in counter.items() if len(values) > 1]


def test_cross_redundancy():
    """Test the redundancy of manually curated mappings and predicted mappings."""
    counter = defaultdict(lambda: defaultdict(list))
    for label, line, mapping in _iter_groups():
        counter[get_canonical_tuple(mapping)][label].append(line)

    redundant = []
    for mapping, label_to_lines in counter.items():
        if len(label_to_lines) <= 1:
            continue
        redundant.append((mapping, sorted(label_to_lines.items())))

    if redundant:
        msg = "".join(
            f"\n  {mapping}: {_locations_str(locations)}" for mapping, locations in redundant
        )
        raise ValueError(f"{len(redundant)} are redundant: {msg}")


def _locations_str(locations):
    return ", ".join(f"{label}:{line}" for label, line in locations)


def _assert_no_internal_redundancies(m, tuple_cls):
    counter = defaultdict(list)
    for line, mapping in enumerate(m, start=1):
        counter[tuple_cls.from_dict(mapping)].append(line)
    redundant = _extract_redundant(counter)
    if redundant:
        msg = "".join(
            f"\n  {mapping.source_curie}/{mapping.target_curie}: {locations}"
            for mapping, locations in redundant
        )
        raise ValueError(f"{len(redundant)} are redundant: {msg}")


def test_predictions_sorted():
    """Test the predictions are in a canonical order."""
    assert predictions == sorted(  # noqa:S101
        predictions, key=mapping_sort_key
    ), "Predictions are not sorted"
    _assert_no_internal_redundancies(predictions, PredictionTuple)


def test_curations_sorted():
    """Test the true curated mappings are in a canonical order."""
    assert mappings == sorted(  # noqa:S101
        mappings, key=mapping_sort_key
    ), "True curations are not sorted"
    _assert_no_internal_redundancies(mappings, MappingTuple)


def test_false_mappings_sorted():
    """Test the false curated mappings are in a canonical order."""
    assert incorrect == sorted(  # noqa:S101
        incorrect, key=mapping_sort_key
    ), "False curations are not sorted"
    _assert_no_internal_redundancies(incorrect, MappingTuple)


def test_unsure_sorted():
    """Test the unsure mappings are in a canonical order."""
    assert unsure == sorted(  # noqa:S101
        unsure, key=mapping_sort_key
    ), "Unsure curations are not sorted"
    _assert_no_internal_redundancies(unsure, MappingTuple)



# -*- coding: utf-8 -*-

"""Export Biomappings as SSSOM."""

import pathlib

import bioregistry
import click
import yaml

from biomappings import load_mappings, load_predictions
from biomappings.utils import DATA, get_curie

DIRECTORY = pathlib.Path(DATA).joinpath("sssom")
DIRECTORY.mkdir(exist_ok=True, parents=True)
TSV_PATH = DIRECTORY.joinpath("biomappings.sssom.tsv")
JSON_PATH = DIRECTORY.joinpath("biomappings.sssom.json")
META_PATH = DIRECTORY.joinpath("biomappings.sssom.yml")

CC0_URL = "https://creativecommons.org/publicdomain/zero/1.0/"
META = {
    "license": CC0_URL,
    "mapping_provider": "https://github.com/biopragmatics/biomappings",
    "mapping_set_group": "biomappings",
    "mapping_set_id": "biomappings",
    "mapping_set_title": "Biomappings",
}
# TODO use this vocabulary directly in file
TYPE_TO_JUSTIFICATION = {
    "lexical": "semapv:LexicalMatching",
    "manual": "semapv:ManualMappingCuration",  # FIXME in source
    "manually_reviewed": "semapv:ManualMappingCuration",
}


def _get_justification(mapping):
    t = mapping["type"]
    if t.startswith("semapv:"):
        return t
    return TYPE_TO_JUSTIFICATION[t]


def get_sssom_df():
    """Get an SSSOM dataframe."""
    import pandas as pd

    creators = set()

    rows = []
    prefixes = set()
    columns = [
        "subject_id",
        "subject_label",
        "predicate_id",
        "object_id",
        "object_label",
        "mapping_justification",
        "author_id",
        "confidence",
        "mapping_tool",
    ]
    for mapping in load_mappings():
        prefixes.add(mapping["source prefix"])
        prefixes.add(mapping["target prefix"])
        source = mapping["source"]
        if any(source.startswith(x) for x in ["orcid:", "wikidata:"]):
            creators.add(source)
        rows.append(
            (
                get_curie(mapping["source prefix"], mapping["source identifier"]),
                mapping["source name"],
                f'{mapping["relation"]}',
                get_curie(mapping["target prefix"], mapping["target identifier"]),
                mapping["target name"],
                _get_justification(mapping),  # match justification
                source,  # curator CURIE
                None,  # no confidence necessary
                None,  # mapping tool: none necessary for manually curated
            )
        )
    for mapping in load_predictions():
        prefixes.add(mapping["source prefix"])
        prefixes.add(mapping["target prefix"])
        rows.append(
            (
                get_curie(mapping["source prefix"], mapping["source identifier"]),
                mapping["source name"],
                f'{mapping["relation"]}',
                get_curie(mapping["target prefix"], mapping["target identifier"]),
                mapping["target name"],
                _get_justification(mapping),  # match justification
                None,  # no curator CURIE
                mapping["confidence"],
                mapping["source"],  # mapping tool: source script
            )
        )
    df = pd.DataFrame(rows, columns=columns)
    return prefixes, sorted(creators), df


@click.command()
def sssom():
    """Export SSSOM."""
    prefixes, creators, df = get_sssom_df()
    df.to_csv(TSV_PATH, sep="\t", index=False)

    # Get a CURIE map containing only the relevant prefixes
    prefix_map = {
        prefix: formatter
        for prefix, formatter in bioregistry.get_prefix_map().items()
        if prefix in prefixes
    }
    with open(META_PATH, "w") as file:
        yaml.safe_dump({"curie_map": prefix_map, "creator_id": creators, **META}, file)

    from sssom.parsers import from_sssom_dataframe
    from sssom.writers import write_json

    try:
        msdf = from_sssom_dataframe(df, prefix_map=prefix_map, meta=META)
    except Exception as e:
        click.secho(f"SSSOM Export failed...\n{e}", fg="red")
        return
    with JSON_PATH.open("w") as file:
        write_json(msdf, file)

    # TODO add RDF export, but it's currently broken in SSSOM-py
    #  and in general the LinkML one is completely unusable (too slow)


if __name__ == "__main__":
    sssom()



# -*- coding: utf-8 -*-

"""Upload the Biomappings to NDEx.

.. seealso:: https://www.ndexbio.org/viewer/networks/402d1fd6-49d6-11eb-9e72-0ac135e8bacf
"""

import bioregistry
import click
import pystow
from tqdm import tqdm

from biomappings import load_mappings
from biomappings.utils import get_curie, get_git_hash

BIOMAPPINGS_NDEX_UUID = "402d1fd6-49d6-11eb-9e72-0ac135e8bacf"


@click.command()
@click.option("--username")
@click.option("--password")
def ndex(username, password):
    """Upload to NDEx."""
    try:
        from ndex2 import NiceCXBuilder
    except ImportError:
        click.secho("Need to `pip install ndex2` before uploading to NDEx", fg="red")
        return

    positive_mappings = load_mappings()
    cx = NiceCXBuilder()
    cx.set_name("Biomappings")
    cx.add_network_attribute(
        "description", "Manually curated mappings (skos:exactMatch) between biological entities."
    )
    cx.add_network_attribute("reference", "https://github.com/biomappings/biomappings")
    cx.add_network_attribute("rights", "Waiver-No rights reserved (CC0)")

    prefixes = {
        prefix
        for mapping in positive_mappings
        for prefix in (mapping["source prefix"], mapping["target prefix"])
    }
    prefixes.add("orcid")
    cx.set_context({prefix: bioregistry.get_uri_prefix(prefix) for prefix in prefixes})

    cx.add_network_attribute("version", get_git_hash())
    authors = sorted(
        set(
            mapping["source"]
            for mapping in positive_mappings
            if mapping["source"].startswith("orcid:")
        )
    )
    cx.add_network_attribute("author", authors, type="list_of_string")

    for mapping in tqdm(positive_mappings, desc="Loading NiceCXBuilder"):
        source = cx.add_node(
            represents=mapping["source name"],
            name=get_curie(mapping["source prefix"], mapping["source identifier"]),
        )
        target = cx.add_node(
            represents=mapping["target name"],
            name=get_curie(mapping["target prefix"], mapping["target identifier"]),
        )
        edge = cx.add_edge(
            source=source,
            target=target,
            interaction=mapping["relation"],
        )
        cx.add_edge_attribute(edge, "type", mapping["type"])
        cx.add_edge_attribute(edge, "provenance", mapping["source"])

    nice_cx = cx.get_nice_cx()
    nice_cx.update_to(
        uuid=BIOMAPPINGS_NDEX_UUID,
        server=pystow.get_config("ndex", "server", default="http://public.ndexbio.org"),
        username=pystow.get_config("ndex", "username", passthrough=username),
        password=pystow.get_config("ndex", "password", passthrough=password),
    )
    click.echo(f"Uploaded to https://bioregistry.io/ndex:{BIOMAPPINGS_NDEX_UUID}")


if __name__ == "__main__":
    ndex()



# -*- coding: utf-8 -*-

"""Utilities."""

import os
import re
from subprocess import CalledProcessError, check_output  # noqa: S404
from typing import Any, Mapping, Optional, Tuple

import bioregistry

HERE = os.path.dirname(os.path.abspath(__file__))
RESOURCE_PATH = os.path.abspath(os.path.join(HERE, "resources"))
DOCS = os.path.abspath(os.path.join(HERE, os.pardir, os.pardir, "docs"))
IMG = os.path.join(DOCS, "img")
DATA = os.path.join(DOCS, "_data")


def get_git_hash() -> Optional[str]:
    """Get the git hash.

    :return:
        The git hash, equals 'UNHASHED' if encountered CalledProcessError, signifying that the
        code is not installed in development mode.
    """
    rv = _git("rev-parse", "HEAD")
    if not rv:
        return None
    return rv[:6]


def commit(message: str) -> Optional[str]:
    """Make a commit with the following message."""
    return _git("commit", "-m", message, "-a")


def push(branch_name: Optional[str] = None) -> Optional[str]:
    """Push the git repo."""
    if branch_name is not None:
        return _git("push", "origin", branch_name)
    else:
        return _git("push")


def not_main() -> bool:
    """Return if on the master branch."""
    return "master" != _git("rev-parse", "--abbrev-ref", "HEAD")


def get_branch() -> str:
    """Return current git branch."""
    rv = _git("branch", "--show-current")
    if rv is None:
        raise RuntimeError
    return rv


def _git(*args: str) -> Optional[str]:
    with open(os.devnull, "w") as devnull:
        try:
            ret = check_output(  # noqa: S603,S607
                ["git", *args],
                cwd=os.path.dirname(__file__),
                stderr=devnull,
            )
        except CalledProcessError as e:
            print(e)  # noqa:T201
            return None
        else:
            return ret.strip().decode("utf-8")


def get_script_url(fname: str) -> str:
    """Get the source path for this script.

    :param fname: Pass ``__file__`` as the argument to this function.
    :return: The script's URL to GitHub
    """
    commit_hash = get_git_hash()
    script_name = os.path.basename(fname)
    return f"https://github.com/biomappings/biomappings/blob/{commit_hash}/scripts/{script_name}"


def get_canonical_tuple(mapping: Mapping[str, Any]) -> Tuple[str, str, str, str]:
    """Get the canonical tuple from a mapping entry."""
    source = mapping["source prefix"], mapping["source identifier"]
    target = mapping["target prefix"], mapping["target identifier"]
    if source > target:
        source, target = target, source
    return (*source, *target)


class UnregisteredPrefix(ValueError):
    """Raised for an invalid prefix."""


class UnstandardizedPrefix(ValueError):
    """Raised for an unstandardized prefix."""

    def __init__(self, prefix: str, norm_prefix: str):
        """Initialize the error.

        :param prefix: A CURIE's prefix
        :param norm_prefix: The normalized prefid
        """
        self.prefix = prefix
        self.norm_prefix = norm_prefix

    def __str__(self) -> str:  # noqa:D105
        return f"{self.prefix} should be standardized to {self.norm_prefix}"


class InvalidIdentifier(ValueError):
    """Raised for an invalid identifier."""

    def __init__(self, prefix: str, identifier: str):
        """Initialize the error.

        :param prefix: A CURIE's prefix
        :param identifier: A CURIE's identifier
        """
        self.prefix = prefix
        self.identifier = identifier


class InvalidIdentifierPattern(InvalidIdentifier):
    """Raised for an identifier that doesn't match the pattern."""

    def __init__(self, prefix: str, identifier: str, pattern):
        """Initialize the error.

        :param prefix: A CURIE's prefix
        :param identifier: A CURIE's identifier
        :param pattern: A regular expression pattern
        """
        super().__init__(prefix, identifier)
        self.pattern = pattern

    def __str__(self) -> str:  # noqa:D105
        return f"{self.prefix}:{self.identifier} does not match pattern {self.pattern}"


class InvalidNormIdentifier(InvalidIdentifier):
    """Raised for an invalid normalized identifier."""

    def __init__(self, prefix: str, identifier: str, norm_identifier: str):
        """Initialize the error.

        :param prefix: A CURIE's prefix
        :param identifier: A CURIE's identifier
        :param norm_identifier: The normalized version of the identifier
        """
        super().__init__(prefix, identifier)
        self.norm_identifier = norm_identifier

    def __str__(self) -> str:  # noqa:D105
        return f"{self.prefix}:{self.identifier} does not match normalized CURIE {self.prefix}:{self.norm_identifier}"


def check_valid_prefix_id(prefix: str, identifier: str):
    """Check the prefix/identifier pair is valid.

    :param prefix:
        The prefix from a CURIE
    :param identifier:
        The local unique identifier from a CURIE
    :raises UnregisteredPrefix:
        if the prefix is not registered with the Bioregistry
    :raises UnstandardizedPrefix:
        if the prefix is not standardized w.r.t. the Bioregistry
    :raises InvalidNormIdentifier:
        if the identifier is not standardized, either against the MIRIAM
        standard, if available, or against the Bioregistry standard
    :raises InvalidIdentifierPattern:
        if the does not match the appropriate regular expression for MIRIAM
        (if available) or for the Bioregistry. If no regular expression is
        available, then this check is not applied.
    :raises RuntimeError:
        If the preconditions for miriam standardization aren't met. However,
        this shouldn't be possible in practice, and this documentation is
        merely a formality.
    """
    resource = bioregistry.get_resource(prefix)
    if resource is None:
        raise UnregisteredPrefix(prefix)
    if prefix != resource.prefix:
        raise UnstandardizedPrefix(prefix, resource.prefix)
    miriam_prefix = resource.get_miriam_prefix()

    # If this resource has a mapping to MIRIAM, the MIRIAM-specific
    # normalization will be applied, which e.g., adds missing
    # redundant prefixes into the local unique identifiers
    if miriam_prefix is not None:
        norm_id = resource.miriam_standardize_identifier(identifier)
        if norm_id is None:
            raise RuntimeError(
                "should not be possible since we check for miriam prefix"
                " before running miriam_standardize_identifier"
            )
        if norm_id != identifier:
            raise InvalidNormIdentifier(prefix, identifier, norm_id)
        if prefix == "pr":
            pattern = None  # identifiers.org is broken for uniprot in PR
        else:
            pattern = re.compile(resource.miriam["pattern"])

    # If this resource does not have a mapping to MIRIAM, then
    # the Bioregistry normalization will be applied, which e.g.,
    # strips potential redundant prefixes in local unique identifiers
    # or any other "bananas"
    else:
        norm_id = resource.standardize_identifier(identifier)
        if norm_id != identifier:
            raise InvalidNormIdentifier(prefix, identifier, norm_id)
        pattern = resource.get_pattern_re()
    if pattern is not None and not pattern.match(identifier):
        raise InvalidIdentifierPattern(prefix, identifier, pattern)


def get_curie(prefix: str, identifier: str) -> str:
    """Get a normalized curie from a pre-parsed prefix/identifier pair."""
    prefix_norm, identifier_norm = bioregistry.normalize_parsed_curie(prefix, identifier)
    if prefix_norm is None or identifier_norm is None:
        raise ValueError(f"could not normalize {prefix}:{identifier}")
    return f"{prefix_norm}:{identifier_norm}"



# -*- coding: utf-8 -*-

"""Generate a summary for the Biomappings website."""

import itertools as itt
import os
import typing
from collections import Counter
from typing import Iterable, Mapping, Optional

import click
import yaml

__all__ = [
    "export",
]


@click.command()
def export():
    """Create export data file."""
    from biomappings.resources import (
        load_false_mappings,
        load_mappings,
        load_predictions,
        load_unsure,
    )
    from biomappings.utils import DATA

    path = os.path.join(DATA, "summary.yml")

    true_mappings = load_mappings()
    false_mappings = load_false_mappings()
    unsure_mappings = load_unsure()
    rv = {
        "positive": _get_counter(true_mappings),
        "negative": _get_counter(false_mappings),
        "unsure": _get_counter(unsure_mappings),
        "predictions": _get_counter(load_predictions()),
        "contributors": _get_contributors(
            itt.chain(true_mappings, false_mappings, unsure_mappings)
        ),
    }
    rv.update(
        {
            f"{k}_mapping_count": sum(e["count"] for e in rv[k])
            for k in ("positive", "negative", "unsure", "predictions")
        }
    )
    rv.update(
        {
            f"{k}_prefix_count": len(
                set(itt.chain.from_iterable((e["source"], e["target"]) for e in rv[k]))
            )
            for k in ("positive", "negative", "unsure", "predictions")
        }
    )
    with open(path, "w") as file:
        yaml.safe_dump(rv, file, indent=2)


def _get_counter(mappings: Iterable[Mapping[str, str]]):
    counter: typing.Counter[typing.Tuple[str, str]] = Counter()
    for mapping in mappings:
        source, target = mapping["source prefix"], mapping["target prefix"]
        if source > target:
            source, target = target, source
        counter[source, target] += 1
    return [
        dict(source=source, target=target, count=count)
        for (source, target), count in counter.most_common()
    ]


def _get_contributors(mappings: Iterable[Mapping[str, str]]):
    from biomappings.resources import load_curators

    curators = {record["orcid"]: record for record in load_curators()}
    counter = Counter(_get_source(mapping["source"]) for mapping in mappings)
    return [
        dict(count=count, **curators[orcid]) if orcid else dict(count=count)
        for orcid, count in counter.most_common()
    ]


def _get_source(source: str) -> Optional[str]:
    if source.startswith("orcid:"):
        return source[len("orcid:") :]
    return None


if __name__ == "__main__":
    export()



# -*- coding: utf-8 -*-

"""The biomappings CLI."""
import sys

import click
from more_click import make_web_command, run_app

from .export_sssom import sssom
from .graph import charts
from .summary import export
from .upload_ndex import ndex
from .utils import get_git_hash


@click.group()
@click.version_option()
def main():
    """Run the biomappings CLI."""


if get_git_hash() is not None:
    # This command is called "web" by default
    main.add_command(make_web_command("biomappings.wsgi:app"))

    @main.command()
    @click.option("--path", required=True, type=click.Path(), help="A predictions TSV file path")
    def curate(path):
        """Run a target curation web app."""
        from .resources import _load_table
        from .wsgi import get_app

        target_curies = []
        for mapping in _load_table(path):
            target_curies.append((mapping["source prefix"], mapping["source identifier"]))
            target_curies.append((mapping["target prefix"], mapping["target identifier"]))
        app = get_app(target_curies=target_curies)
        run_app(app, with_gunicorn=False)

else:

    @main.command()
    def web():
        """Show an error for the web interface."""
        click.secho(
            "You are not running biomappings from a development installation.\n"
            "Please run the following to install in development mode:\n"
            "  $ git clone https://github.com/biomappings/biomappings.git\n"
            "  $ cd biomappings\n"
            "  $ pip install -e .",
            fg="red",
        )


@main.command()
@click.pass_context
def update(ctx: click.Context):
    """Run all update functions."""
    click.secho("Building general exports", fg="green")
    ctx.invoke(export)
    click.secho("Building SSSOM export", fg="green")
    ctx.invoke(sssom)
    click.secho("Generating charts", fg="green")
    ctx.invoke(charts)
    # click.secho("Uploading to NDEx", fg="green")
    # ctx.invoke(ndex)


@main.command()
@click.option("--standardize", is_flag=True)
def lint(standardize: bool):
    """Sort files and remove duplicates."""
    from .resources import (
        lint_false_mappings,
        lint_predictions,
        lint_true_mappings,
        lint_unsure_mappings,
    )

    lint_true_mappings()
    lint_false_mappings()
    lint_unsure_mappings()
    lint_predictions(standardize=standardize)


@main.command()
@click.argument("prefixes", nargs=-1)
def prune(prefixes):
    """Prune inferred mappings between the given prefixes from the predictions."""
    from .mapping_graph import get_custom_filter
    from .resources import filter_predictions

    if len(prefixes) < 2:
        click.secho("Must give at least 2 prefixes", fg="red")
        return sys.exit(0)

    cf = get_custom_filter(prefixes[0], prefixes[1:])
    filter_predictions(cf)


@main.command()
def remove_curated():
    """Remove curated mappings from the predicted mappings, use this if they get out of sync."""
    from .resources import filter_predictions, get_curated_filter

    filter_predictions(get_curated_filter())


main.add_command(export)
main.add_command(ndex)
main.add_command(charts)
main.add_command(sssom)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Biomappings resources."""

import csv
import itertools as itt
import os
from collections import defaultdict
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)

import bioregistry
from tqdm import tqdm

from biomappings.utils import RESOURCE_PATH, get_canonical_tuple

MAPPINGS_HEADER = [
    "source prefix",
    "source identifier",
    "source name",
    "relation",
    "target prefix",
    "target identifier",
    "target name",
    "type",
    "source",
    "prediction_type",
    "prediction_source",
    "prediction_confidence",
]
PREDICTIONS_HEADER = [
    "source prefix",
    "source identifier",
    "source name",
    "relation",
    "target prefix",
    "target identifier",
    "target name",
    "type",
    "confidence",
    "source",
]


class MappingTuple(NamedTuple):
    """A named tuple class for mappings."""

    source_prefix: str
    source_id: str
    source_name: str
    relation: str
    target_prefix: str
    target_identifier: str
    target_name: str
    type: str
    source: str
    prediction_type: Optional[str]
    prediction_source: Optional[str]
    prediction_confidence: Optional[float]

    def as_dict(self) -> Mapping[str, Any]:
        """Get the mapping tuple as a dictionary."""
        return dict(zip(MAPPINGS_HEADER, self))  # type:ignore

    @classmethod
    def from_dict(cls, mapping: Mapping[str, str]) -> "MappingTuple":
        """Get the mapping tuple from a dictionary."""
        values = []
        for key in MAPPINGS_HEADER:
            value = mapping.get(key) or None
            if key == "prediction_confidence" and value is not None:
                value = float(value)  # type:ignore
            values.append(value)
        return cls(*values)  # type:ignore

    @property
    def source_curie(self) -> str:
        """Concatenate the source prefix and ID to a CURIE."""
        return f"{self.source_prefix}:{self.source_id}"

    @property
    def target_curie(self) -> str:
        """Concatenate the target prefix and ID to a CURIE."""
        return f"{self.target_prefix}:{self.target_identifier}"


class PredictionTuple(NamedTuple):
    """A named tuple class for predictions."""

    source_prefix: str
    source_id: str
    source_name: str
    relation: str
    target_prefix: str
    target_identifier: str
    target_name: str
    type: str
    confidence: float
    source: str

    def as_dict(self) -> Mapping[str, Any]:
        """Get the prediction tuple as a dictionary."""
        return dict(zip(PREDICTIONS_HEADER, self))  # type:ignore

    @classmethod
    def from_dict(cls, mapping: Mapping[str, str]) -> "PredictionTuple":
        """Get the prediction tuple from a dictionary."""
        values = []
        for key in PREDICTIONS_HEADER:
            value = mapping.get(key) or None
            if key == "confidence" and value is not None:
                value = float(value)  # type:ignore
            values.append(value)
        return cls(*values)  # type:ignore

    @property
    def source_curie(self) -> str:
        """Concatenate the source prefix and ID to a CURIE."""
        return f"{self.source_prefix}:{self.source_id}"

    @property
    def target_curie(self) -> str:
        """Concatenate the target prefix and ID to a CURIE."""
        return f"{self.target_prefix}:{self.target_identifier}"


def get_resource_file_path(fname) -> str:
    """Get a resource by its file name."""
    return os.path.join(RESOURCE_PATH, fname)


def _load_table(fname) -> List[Dict[str, str]]:
    with open(fname, "r") as fh:
        reader = csv.reader(fh, delimiter="\t")
        header = next(reader)
        return [_clean(header, row) for row in reader]


def _clean(header, row):
    d = dict(zip(header, row))
    return {k: v if v else None for k, v in d.items()}


def _write_helper(
    header: Sequence[str], lod: Iterable[Mapping[str, str]], path: str, mode: str
) -> None:
    lod = sorted(lod, key=mapping_sort_key)
    with open(path, mode) as file:
        if mode == "w":
            print(*header, sep="\t", file=file)  # noqa:T201
        for line in lod:
            print(*[line[k] or "" for k in header], sep="\t", file=file)  # noqa:T201


def mapping_sort_key(prediction: Mapping[str, str]) -> Tuple[str, ...]:
    """Return a tuple for sorting mapping dictionaries."""
    return (
        prediction["source prefix"],
        prediction["source identifier"],
        prediction["relation"],
        prediction["target prefix"],
        prediction["target identifier"],
        prediction["type"],
        prediction["source"] or "",
    )


TRUE_MAPPINGS_PATH = get_resource_file_path("mappings.tsv")


def load_mappings() -> List[Dict[str, str]]:
    """Load the mappings table."""
    return _load_table(TRUE_MAPPINGS_PATH)


def append_true_mappings(m: Iterable[Mapping[str, str]], sort: bool = True) -> None:
    """Append new lines to the mappings table."""
    _write_helper(MAPPINGS_HEADER, m, TRUE_MAPPINGS_PATH, "a")
    if sort:
        lint_true_mappings()


def append_true_mapping_tuples(mappings: Iterable[MappingTuple]) -> None:
    """Append new lines to the mappings table."""
    append_true_mappings(mapping.as_dict() for mapping in set(mappings))


def write_true_mappings(m: Iterable[Mapping[str, str]]) -> None:
    """Write mappigns to the true mappings file."""
    _write_helper(MAPPINGS_HEADER, m, TRUE_MAPPINGS_PATH, "w")


def lint_true_mappings() -> None:
    """Lint the true mappings file."""
    mappings = load_mappings()
    mappings = _remove_redundant(mappings, MappingTuple)
    write_true_mappings(sorted(mappings, key=mapping_sort_key))


FALSE_MAPPINGS_PATH = get_resource_file_path("incorrect.tsv")


def load_false_mappings() -> List[Dict[str, str]]:
    """Load the false mappings table."""
    return _load_table(FALSE_MAPPINGS_PATH)


def append_false_mappings(m: Iterable[Mapping[str, str]], sort: bool = True) -> None:
    """Append new lines to the false mappings table."""
    _write_helper(MAPPINGS_HEADER, m, FALSE_MAPPINGS_PATH, "a")
    if sort:
        lint_false_mappings()


def write_false_mappings(m: Iterable[Mapping[str, str]]) -> None:
    """Write mappings to the false mappings file."""
    _write_helper(MAPPINGS_HEADER, m, FALSE_MAPPINGS_PATH, "w")


def lint_false_mappings() -> None:
    """Lint the false mappings file."""
    mappings = load_false_mappings()
    mappings = _remove_redundant(mappings, MappingTuple)
    write_false_mappings(sorted(mappings, key=mapping_sort_key))


UNSURE_PATH = get_resource_file_path("unsure.tsv")


def load_unsure() -> List[Dict[str, str]]:
    """Load the unsure table."""
    return _load_table(UNSURE_PATH)


def append_unsure_mappings(m: Iterable[Mapping[str, str]], sort: bool = True) -> None:
    """Append new lines to the "unsure" mappings table."""
    _write_helper(MAPPINGS_HEADER, m, UNSURE_PATH, "a")
    if sort:
        lint_unsure_mappings()


def write_unsure_mappings(m: Iterable[Mapping[str, str]]) -> None:
    """Write mappings to the unsure mappings file."""
    _write_helper(MAPPINGS_HEADER, m, UNSURE_PATH, "w")


def lint_unsure_mappings() -> None:
    """Lint the unsure mappings file."""
    mappings = load_unsure()
    mappings = _remove_redundant(mappings, MappingTuple)
    write_unsure_mappings(sorted(mappings, key=mapping_sort_key))


PREDICTIONS_PATH = get_resource_file_path("predictions.tsv")


def load_predictions() -> List[Dict[str, str]]:
    """Load the predictions table."""
    return _load_table(PREDICTIONS_PATH)


def write_predictions(m: Iterable[Mapping[str, str]]) -> None:
    """Write new content to the predictions table."""
    _write_helper(PREDICTIONS_HEADER, m, PREDICTIONS_PATH, "w")


def append_prediction_tuples(
    prediction_tuples: Iterable[PredictionTuple], deduplicate: bool = True, sort: bool = True
) -> None:
    """Append new lines to the predictions table that come as canonical tuples."""
    append_predictions(
        (prediction_tuple.as_dict() for prediction_tuple in set(prediction_tuples)),
        deduplicate=deduplicate,
        sort=sort,
    )


def append_predictions(
    mappings: Iterable[Mapping[str, str]], deduplicate: bool = True, sort: bool = True
) -> None:
    """Append new lines to the predictions table."""
    if deduplicate:
        existing_mappings = {
            get_canonical_tuple(existing_mapping)
            for existing_mapping in itt.chain(
                load_mappings(),
                load_false_mappings(),
                load_unsure(),
                load_predictions(),
            )
        }
        mappings = (
            mapping for mapping in mappings if get_canonical_tuple(mapping) not in existing_mappings
        )

    _write_helper(PREDICTIONS_HEADER, mappings, PREDICTIONS_PATH, "a")
    if sort:
        lint_predictions()


def lint_predictions(standardize: bool = False) -> None:
    """Lint the predictions file.

    1. Make sure there are no redundant rows
    2. Make sure no rows in predictions match a row in the curated files
    3. Make sure it's sorted

    :param standardize: Should identifiers be standardized (against the
             combination of Identifiers.org and Bioregistry)?
    """
    curated_mappings = {
        get_canonical_tuple(mapping)
        for mapping in itt.chain(
            load_mappings(),
            load_false_mappings(),
            load_unsure(),
        )
    }
    mappings = [
        mapping
        for mapping in tqdm(
            load_predictions(), desc="Removing curated from predicted", unit_scale=True
        )
        if get_canonical_tuple(mapping) not in curated_mappings
    ]
    mappings = _remove_redundant(mappings, PredictionTuple, standardize=standardize)
    write_predictions(sorted(mappings, key=mapping_sort_key))


def _remove_redundant(mappings, tuple_cls, standardize: bool = False):
    if standardize:
        mappings = (
            _standardize_mapping(mapping)
            for mapping in tqdm(mappings, desc="Standardizing mappings", unit_scale=True)
        )
    return (mapping.as_dict() for mapping in {tuple_cls.from_dict(mapping) for mapping in mappings})


def _standardize_mapping(mapping):
    """Standardize a mapping."""
    for prefix_key, identifier_key in [
        ("source prefix", "source identifier"),
        ("target prefix", "target identifier"),
    ]:
        prefix, identifier = mapping[prefix_key], mapping[identifier_key]
        resource = bioregistry.get_resource(prefix)
        if resource is None:
            raise ValueError
        miriam_prefix = resource.get_miriam_prefix()
        if miriam_prefix is not None:
            mapping[identifier_key] = (
                resource.miriam_standardize_identifier(identifier) or identifier
            )
        else:
            mapping[identifier_key] = resource.standardize_identifier(identifier)
    return mapping


def load_curators():
    """Load the curators table."""
    return _load_table(get_resource_file_path("curators.tsv"))


def filter_predictions(custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]]) -> None:
    """Filter all of the predictions by removing what's in the custom filter then re-write.

    :param custom_filter: A filter 3-dictionary of source prefix to target prefix
        to source identifier to target identifier
    """
    predictions = load_predictions()
    predictions = [
        prediction for prediction in predictions if _check_filter(prediction, custom_filter)
    ]
    write_predictions(predictions)


def _check_filter(
    prediction: Mapping[str, str],
    custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]],
) -> bool:
    source_prefix, target_prefix = prediction["source prefix"], prediction["target prefix"]
    source_id, target_id = prediction["source identifier"], prediction["target identifier"]
    return target_id != custom_filter.get(source_prefix, {}).get(target_prefix, {}).get(source_id)


def get_curated_filter() -> Mapping[str, Mapping[str, Mapping[str, str]]]:
    """Get a filter over all curated mappings."""
    d: DefaultDict[str, DefaultDict[str, Dict[str, str]]] = defaultdict(lambda: defaultdict(dict))
    for m in itt.chain(load_mappings(), load_false_mappings(), load_unsure()):
        d[m["source prefix"]][m["target prefix"]][m["source identifier"]] = m["target identifier"]
    return {k: dict(v) for k, v in d.items()}



# -*- coding: utf-8 -*-

"""Functions for working with the mapping graph."""

import itertools as itt
from collections import defaultdict
from typing import DefaultDict, Dict, Iterable, Mapping, Optional

import networkx as nx
import pyobo


def get_custom_filter(
    prefix: str, targets: Iterable[str]
) -> Mapping[str, Mapping[str, Mapping[str, str]]]:
    """Get a custom filter dictionary induced over the mutual mapping graph with all target prefixes.

    :param prefix: The source prefix
    :param targets: All potential target prefixes
    :returns: A filter 3-dictionary of source prefix to target prefix to source identifier to target identifier
    """
    graph = mutual_mapping_graph([prefix, *targets])
    rv: DefaultDict[str, Dict[str, str]] = defaultdict(dict)
    for p, identifier in graph:
        if p != prefix:
            continue
        for xref_prefix, xref_identifier in nx.single_source_shortest_path(graph, (p, identifier)):
            rv[xref_prefix][identifier] = xref_identifier
    return {prefix: dict(rv)}


def mutual_mapping_graph(
    prefixes: Iterable[str],
    skip_sources: Optional[Iterable[str]] = None,
    skip_targets: Optional[Iterable[str]] = None,
) -> nx.Graph:
    """Get the undirected mapping graph between the given prefixes.

    :param prefixes: A list of prefixes to use with :func:`pyobo.get_filtered_xrefs` to get xrefs.
    :param skip_sources: An optional list of prefixes to skip as the source for xrefs
    :param skip_targets: An optional list of prefixes to skip as the target for xrefs
    :return: The undirected mapping graph containing mappings between entries in the given namespaces.
    """
    prefixes = sorted(prefixes)
    skip_sources = set() if skip_sources is None else set(skip_sources)
    skip_targets = set() if skip_targets is None else set(skip_targets)
    graph = nx.Graph()
    for source, target in itt.product(prefixes, repeat=2):
        if source == target or source in skip_sources or target in skip_targets:
            continue
        for source_id, target_id in pyobo.get_filtered_xrefs(source, target).items():
            graph.add_edge((source, source_id), (target, target_id))
    return graph



# -*- coding: utf-8 -*-

"""Web curation interface for :mod:`biomappings`."""

import getpass
import os
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union

import bioregistry
import flask
import flask_bootstrap
from flask import current_app
from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField

from biomappings.resources import (
    append_false_mappings,
    append_true_mappings,
    append_unsure_mappings,
    load_curators,
    load_predictions,
    write_predictions,
)
from biomappings.utils import (
    check_valid_prefix_id,
    commit,
    get_branch,
    get_curie,
    not_main,
    push,
)


def get_app(target_curies: Optional[Iterable[Tuple[str, str]]] = None) -> flask.Flask:
    """Get a curation flask app."""
    app_ = flask.Flask(__name__)
    app_.config["WTF_CSRF_ENABLED"] = False
    app_.config["SECRET_KEY"] = os.urandom(8)
    app_.config["SHOW_RELATIONS"] = True
    app_.config["SHOW_LINES"] = False
    controller = Controller(target_curies=target_curies)
    app_.config["controller"] = controller
    flask_bootstrap.Bootstrap4(app_)
    app_.register_blueprint(blueprint)
    return app_


# A mapping from your computer's user, returned by getuser.getpass()
KNOWN_USERS = {record["user"]: record["orcid"] for record in load_curators()}


def _manual_source():
    known_user = KNOWN_USERS.get(getpass.getuser())
    if known_user:
        return f"orcid:{known_user}"
    return "web"


class Controller:
    """A module for interacting with the predictions and mappings."""

    def __init__(self, target_curies: Optional[Iterable[Tuple[str, str]]] = None):
        """Instantiate the web controller.

        :param target_curies: Pairs of prefix, local unique identifiers that are the target
            of curation. If this is given, pre-filters will be made before on predictions
            to only show ones where either the source or target appears in this set
        """
        self._predictions = load_predictions()
        self._marked: Dict[int, str] = {}
        self.total_curated = 0
        self._added_mappings: List[Dict[str, Union[None, str, float]]] = []
        self.target_ids = set(target_curies or [])

    def predictions(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        query: Optional[str] = None,
        source_query: Optional[str] = None,
        source_prefix: Optional[str] = None,
        target_query: Optional[str] = None,
        target_prefix: Optional[str] = None,
        prefix: Optional[str] = None,
        sort: Optional[str] = None,
        same_text: bool = False,
        provenance: Optional[str] = None,
    ) -> Iterable[Tuple[int, Mapping[str, Any]]]:
        """Iterate over predictions.

        :param offset: If given, offset the iteration by this number
        :param limit: If given, only iterate this number of predictions.
        :param query: If given, show only equivalences that have it appearing as a substring in one of the source
            or target fields.
        :param source_query: If given, show only equivalences that have it appearing as a substring in one of the source
            fields.
        :param source_prefix: If given, show only mappings that have it appearing in the source prefix field
        :param target_query: If given, show only equivalences that have it appearing as a substring in one of the target
            fields.
        :param target_prefix: If given, show only mappings that have it appearing in the target prefix field
        :param prefix: If given, show only equivalences that have it appearing as a substring in one of the prefixes.
        :param same_text: If true, filter to predictions with the same label
        :param sort: If "desc", sorts in descending confidence order. If "asc", sorts in increasing confidence order.
            Otherwise, do not sort.
        :param provenance: If given, filters to provenance values matching this
        :yields: Pairs of positions and prediction dictionaries
        """
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        if offset is not None:
            try:
                for _ in range(offset):
                    next(it)
            except StopIteration:
                # if next() fails, then there are no remaining entries.
                # do not pass go, do not collect 200 euro $
                return
        if limit is None:
            yield from it
        else:
            for line_prediction, _ in zip(it, range(limit)):
                yield line_prediction

    def count_predictions(
        self,
        query: Optional[str] = None,
        source_query: Optional[str] = None,
        source_prefix: Optional[str] = None,
        target_query: Optional[str] = None,
        target_prefix: Optional[str] = None,
        prefix: Optional[str] = None,
        sort: Optional[str] = None,
        same_text: bool = False,
        provenance: Optional[str] = None,
    ) -> int:
        """Count the number of predictions to check for the given filters."""
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        return sum(1 for _ in it)

    def _help_it_predictions(
        self,
        query: Optional[str] = None,
        source_query: Optional[str] = None,
        source_prefix: Optional[str] = None,
        target_query: Optional[str] = None,
        target_prefix: Optional[str] = None,
        prefix: Optional[str] = None,
        sort: Optional[str] = None,
        same_text: bool = False,
        provenance: Optional[str] = None,
    ):
        it: Iterable[Tuple[int, Mapping[str, Any]]] = enumerate(self._predictions)
        if self.target_ids:
            it = (
                (line, p)
                for (line, p) in it
                if (p["source prefix"], p["source identifier"]) in self.target_ids
                or (p["target prefix"], p["target identifier"]) in self.target_ids
            )

        if query is not None:
            it = self._help_filter(
                query,
                it,
                {
                    "source prefix",
                    "source identifier",
                    "source name",
                    "target prefix",
                    "target identifier",
                    "target name",
                },
            )
        if source_prefix is not None:
            it = self._help_filter(source_prefix, it, {"source prefix"})
        if source_query is not None:
            it = self._help_filter(
                source_query, it, {"source prefix", "source identifier", "source name"}
            )
        if target_query is not None:
            it = self._help_filter(
                target_query, it, {"target prefix", "target identifier", "target name"}
            )
        if target_prefix is not None:
            it = self._help_filter(target_prefix, it, {"target prefix"})
        if prefix is not None:
            it = self._help_filter(prefix, it, {"source prefix", "target prefix"})
        if provenance is not None:
            it = self._help_filter(provenance, it, {"source"})

        if sort is not None:
            it = iter(sorted(it, key=lambda l_p: l_p[1]["confidence"], reverse=sort == "desc"))

        if same_text:
            it = (
                (line, prediction)
                for line, prediction in it
                if prediction["source name"].casefold() == prediction["target name"].casefold()
                and prediction["relation"] == "skos:exactMatch"
            )

        rv = ((line, prediction) for line, prediction in it if line not in self._marked)
        return rv

    @staticmethod
    def _help_filter(query: str, it, elements: Set[str]):
        query = query.casefold()
        return (
            (line, prediction)
            for line, prediction in it
            if any(query in prediction[element].casefold() for element in elements)
        )

    @staticmethod
    def get_curie(prefix: str, identifier: str) -> str:
        """Return CURIE for a given prefix and identifier."""
        return get_curie(prefix, identifier)

    @classmethod
    def get_url(cls, prefix: str, identifier: str) -> str:
        """Return URL for a given prefix and identifier."""
        if bioregistry.get_obofoundry_prefix(prefix):
            return bioregistry.get_ols_iri(prefix, identifier)
        return bioregistry.get_bioregistry_iri(prefix, identifier)

    @property
    def total_predictions(self) -> int:
        """Return the total number of yet unmarked predictions."""
        return len(self._predictions) - len(self._marked)

    def mark(self, line: int, value: str) -> None:
        """Mark the given equivalency as correct.

        :param line: Position of the prediction
        :param value: Value to mark the prediction with
        :raises ValueError: if an invalid value is used
        """
        if line not in self._marked:
            self.total_curated += 1
        if value not in {"correct", "incorrect", "unsure"}:
            raise ValueError
        self._marked[line] = value

    def add_mapping(
        self,
        source_prefix: str,
        source_id: str,
        source_name: str,
        target_prefix: str,
        target_id: str,
        target_name: str,
    ) -> None:
        """Add manually curated new mappings."""
        try:
            check_valid_prefix_id(source_prefix, source_id)
        except ValueError as e:
            flask.flash(
                f"Problem with source CURIE {source_prefix}:{source_id}: {e.__class__.__name__}",
                category="warning",
            )
            return

        try:
            check_valid_prefix_id(target_prefix, target_id)
        except ValueError as e:
            flask.flash(
                f"Problem with target CURIE {target_prefix}:{target_id}: {e.__class__.__name__}",
                category="warning",
            )
            return

        self._added_mappings.append(
            {
                "source prefix": source_prefix,
                "source identifier": source_id,
                "source name": source_name,
                "relation": "skos:exactMatch",
                "target prefix": target_prefix,
                "target identifier": target_id,
                "target name": target_name,
                "source": _manual_source(),
                "type": "manual",
                "prediction_type": None,
                "prediction_source": None,
                "prediction_confidence": None,
            }
        )
        self.total_curated += 1

    def persist(self):
        """Save the current markings to the source files."""
        entries = defaultdict(list)

        for line, value in sorted(self._marked.items(), reverse=True):
            prediction = self._predictions.pop(line)
            prediction["prediction_type"] = prediction.pop("type")
            prediction["prediction_source"] = prediction.pop("source")
            prediction["prediction_confidence"] = prediction.pop("confidence")
            prediction["source"] = _manual_source()
            prediction["type"] = "manually_reviewed"
            entries[value].append(prediction)

        append_true_mappings(entries["correct"])
        append_false_mappings(entries["incorrect"])
        append_unsure_mappings(entries["unsure"])
        write_predictions(self._predictions)
        self._marked.clear()

        # Now add manually curated mappings
        append_true_mappings(self._added_mappings)
        self._added_mappings = []


class MappingForm(FlaskForm):
    """Form for entering new mappings."""

    source_prefix = StringField("Source Prefix", id="source_prefix")
    source_id = StringField("Source ID", id="source_id")
    source_name = StringField("Source Name", id="source_name")
    target_prefix = StringField("Target Prefix", id="target_prefix")
    target_id = StringField("Target ID", id="target_id")
    target_name = StringField("Target Name", id="target_name")
    submit = SubmitField("Add")


blueprint = flask.Blueprint("ui", __name__)


@blueprint.route("/")
def home():
    """Serve the home page."""
    form = MappingForm()
    limit = flask.request.args.get("limit", type=int, default=10)
    offset = flask.request.args.get("offset", type=int, default=0)
    query = flask.request.args.get("query")
    source_query = flask.request.args.get("source_query")
    source_prefix = flask.request.args.get("source_prefix")
    target_query = flask.request.args.get("target_query")
    target_prefix = flask.request.args.get("target_prefix")
    provenance = flask.request.args.get("provenance")
    prefix = flask.request.args.get("prefix")
    sort = flask.request.args.get("sort")
    same_text = flask.request.args.get("same_text", default="false").lower() in {"true", "t"}
    show_relations = current_app.config["SHOW_RELATIONS"]
    show_lines = current_app.config["SHOW_LINES"]
    return flask.render_template(
        "home.html",
        controller=current_app.config["controller"],
        form=form,
        limit=limit,
        offset=offset,
        query=query,
        source_query=source_query,
        source_prefix=source_prefix,
        target_query=target_query,
        target_prefix=target_prefix,
        prefix=prefix,
        sort=sort,
        same_text=same_text,
        provenance=provenance,
        # configured
        show_relations=show_relations,
        show_lines=show_lines,
    )


@blueprint.route("/add_mapping", methods=["POST"])
def add_mapping():
    """Add a new mapping manually."""
    form = MappingForm()
    if form.is_submitted():
        current_app.config["controller"].add_mapping(
            form.data["source_prefix"],
            form.data["source_id"],
            form.data["source_name"],
            form.data["target_prefix"],
            form.data["target_id"],
            form.data["target_name"],
        )
        current_app.config["controller"].persist()
    else:
        flask.flash("missing form data", category="warning")
    return _go_home()


@blueprint.route("/commit")
def run_commit():
    """Make a commit then redirect to the the home page."""
    commit_info = commit(
        f'Curated {current_app.config["controller"].total_curated} mapping'
        f'{"s" if current_app.config["controller"].total_curated > 1 else ""}'
        f" ({getpass.getuser()})",
    )
    current_app.logger.warning("git commit res: %s", commit_info)
    if not_main():
        branch = get_branch()
        push_output = push(branch_name=branch)
        current_app.logger.warning("git push res: %s", push_output)
    else:
        flask.flash("did not push because on master branch")
        current_app.logger.warning("did not push because on master branch")
    current_app.config["controller"].total_curated = 0
    return _go_home()


CORRECT = {"yup", "true", "t", "correct", "right", "close enough", "disco"}
INCORRECT = {"no", "nope", "false", "f", "nada", "nein", "incorrect", "negative", "negatory"}
UNSURE = {"unsure", "maybe", "idk", "idgaf", "idgaff"}


def _normalize_mark(value: str) -> str:
    value = value.lower()
    if value in CORRECT:
        return "correct"
    elif value in INCORRECT:
        return "incorrect"
    elif value in UNSURE:
        return "unsure"
    else:
        raise ValueError


@blueprint.route("/mark/<int:line>/<value>")
def mark(line: int, value: str):
    """Mark the given line as correct or not."""
    current_app.config["controller"].mark(line, _normalize_mark(value))
    current_app.config["controller"].persist()
    return _go_home()


def _go_home():
    return flask.redirect(
        flask.url_for(
            ".home",
            limit=flask.request.args.get("limit", type=int),
            offset=flask.request.args.get("offset", type=int),
            query=flask.request.args.get("query"),
            source_query=flask.request.args.get("source_query"),
            source_prefix=flask.request.args.get("source_prefix"),
            target_query=flask.request.args.get("target_query"),
            target_prefix=flask.request.args.get("target_prefix"),
            prefix=flask.request.args.get("prefix"),
            same_text=flask.request.args.get("same_text", default="false").lower() in {"true", "t"},
            provenance=flask.request.args.get("provenance"),
            sort=flask.request.args.get("sort"),
            # config
            show_relations=current_app.config["SHOW_RELATIONS"],
            show_lines=current_app.config["SHOW_LINES"],
        )
    )


app = get_app()

if __name__ == "__main__":
    app.run()
